from collections import Counter

from valanga.representation_factory import RepresentationFactory


class DummyRepresentation:
    pass


def test_create_from_transition_root_uses_full_creator():
    calls = Counter()

    def create_from_state(state):
        calls["create_from_state"] += 1
        return ("state", state)

    factory = RepresentationFactory(
        create_from_state=create_from_state, create_from_state_and_modifications=None
    )

    result = factory.create_from_transition(
        state="root", previous_state_representation=None, modifications=None
    )

    assert result == ("state", "root")
    assert calls["create_from_state"] == 1


def test_create_from_transition_no_modifications_falls_back_to_state():
    calls = Counter()

    def create_from_state(state):
        calls["create_from_state"] += 1
        return ("state", state)

    factory = RepresentationFactory(
        create_from_state=create_from_state, create_from_state_and_modifications=None
    )

    result = factory.create_from_transition(
        state="child",
        previous_state_representation=DummyRepresentation(),
        modifications=None,
    )

    assert result == ("state", "child")
    assert calls["create_from_state"] == 1


def test_create_from_transition_with_modifications():
    calls = Counter()

    def create_from_state_and_modifications(
        state, state_modifications, previous_state_representation
    ):
        calls["create_from_state_and_modifications"] += 1
        return (state, state_modifications, previous_state_representation)

    factory = RepresentationFactory(
        create_from_state=lambda state: ("unused", state),
        create_from_state_and_modifications=create_from_state_and_modifications,
    )

    previous = DummyRepresentation()
    result = factory.create_from_transition(
        state="child", previous_state_representation=previous, modifications="delta"
    )

    assert result == ("child", "delta", previous)
    assert calls["create_from_state_and_modifications"] == 1



"""
Factory for creating content representations from game states and state modifications.
"""

from dataclasses import dataclass
from typing import Protocol

from .game import State
from .represention_for_evaluation import ContentRepresentation


class CreateFromState[StateT: State, EvalIn](Protocol):
    """
    Protocol for creating a state representation from a state.
    """

    def __call__(self, state: StateT) -> ContentRepresentation[StateT, EvalIn]: ...


class CreateFromStateAndModifications[StateT: State, EvalIn, StateModT](Protocol):
    """
    Protocol for creating a state representation from a state and modifications.
    """

    def __call__(
        self,
        state: StateT,
        state_modifications: StateModT,
        previous_state_representation: ContentRepresentation[StateT, EvalIn],
    ) -> ContentRepresentation[StateT, EvalIn]: ...


@dataclass
class RepresentationFactory[StateT: State, EvalIn, StateModT]:
    """Factory for creating content representations from states and state modifications.
    Attributes:
        create_from_state: Function to create a content representation from a state.
        create_from_state_and_modifications: Function to create a content representation from a state and state modifications.
    """

    create_from_state: CreateFromState[StateT, EvalIn]
    create_from_state_and_modifications: CreateFromStateAndModifications[
        StateT, EvalIn, StateModT
    ]

    def create_from_transition(
        self,
        state: StateT,
        previous_state_representation: ContentRepresentation[StateT, EvalIn] | None,
        modifications: StateModT | None,
    ) -> ContentRepresentation[StateT, EvalIn]:
        """Creates a content representation from a state transition.
        Args:
            state: The current state of the game.
            previous_state_representation: The content representation of the previous state, or None if not available.
            modifications: The modifications applied to the previous state to reach the current state, or None if not available.
        Returns:
            ContentRepresentation[StateT, EvalIn]: The content representation of the current state.
        """
        if previous_state_representation is None or modifications is None:
            return self.create_from_state(state)
        return self.create_from_state_and_modifications(
            state, modifications, previous_state_representation
        )



"""Small shared types used by evaluators and representations.

This module exists to avoid runtime circular imports between `evaluations` and
`represention_for_evaluation`.
"""

from typing import Annotated

EvaluatorInput = Annotated[
    object, "The input type for the evaluator, typically a tensor or array"
]

__all__ = ["EvaluatorInput"]



"""Placeholder test file."""


def test_package_imports():
    """Verify the package can be imported."""
    import valanga

    assert valanga is not None


def test_placeholder():
    """Placeholder test - remove when real tests are added."""
    assert True



"""
Common types and utilities representing game objects shared by multiple libraries.
"""

from collections.abc import Hashable
from dataclasses import dataclass, field
from enum import Enum
from typing import Annotated, Any, Iterator, Protocol, Self, Sequence, TypeVar

type Seed = Annotated[int, "seed"]

type StateTag = Annotated[Hashable, "A label or identifier for a state in a game"]

type StateModifications = Annotated[
    object, "Modifications to the state between to time steps of the game"
]  # used for time and memory optimisation

type BranchName = Annotated[str, "A human-readable name for a branch of a state"]

type BranchKey = Annotated[Hashable, "A label or identifier for a branch in a tree"]
type ActionName = Annotated[str, "A human-readable name for an action of a state"]

type ActionKey = Annotated[Hashable, "A label or identifier for an action"]

T_co = TypeVar("T_co", bound=Hashable, covariant=True)


class BranchKeyGeneratorP(Protocol[T_co]):
    """Protocol for a branch key generator that yields branch keys."""

    # whether to sort the branch keys by their respective uci for easy comparison of various implementations
    sort_branch_keys: bool = False

    @property
    def all_generated_keys(self) -> Sequence[T_co] | None:
        """Returns all generated branch keys if available, otherwise None."""
        ...

    def __iter__(self) -> Iterator[T_co]:
        """Returns an iterator over the branch keys."""
        ...

    def __next__(self) -> T_co:
        """Returns the next branch key."""
        ...

    def more_than_one(self) -> bool:
        """Checks if there is more than one branch available.

        Returns:
            bool: True if there is more than one branch, False otherwise.
        """
        ...

    def get_all(self) -> Sequence[T_co]:
        """Returns a list of all branch keys."""
        ...

    def copy_with_reset(self) -> Self:
        """Creates a copy of the legal move generator with an optional reset of generated moves.

        Returns:
            Self: A new instance of the legal move generator with the specified generated moves.
        """
        ...


class State(Protocol):
    """Protocol for a content object that has a tag."""

    @property
    def tag(self) -> StateTag:
        """Returns the tag of the content.

        Returns:
            StateTag: The tag of the content.
        """
        ...

    @property
    def branch_keys(self) -> BranchKeyGeneratorP[BranchKey]:
        """Returns the branch keys associated with the content.

        Returns:
            BranchKeyGeneratorP: The branch keys associated with the content.
        """
        ...

    def branch_name_from_key(self, key: BranchKey) -> str:
        """Returns the branch name corresponding to the given branch key.

        Args:
            key (BranchKey): The branch key.

        Returns:
            str: The branch name corresponding to the given branch key.
        """
        ...

    def branch_key_from_name(self, name: str) -> BranchKey:
        """Returns the branch key corresponding to the given branch name.

        Args:
            name (str): The branch name.
        Returns:
            BranchKey: The branch key corresponding to the given branch name.
        """
        ...

    def is_game_over(self) -> bool:
        """Checks if the game represented by the content is over.

        Returns:
            bool: True if the game is over, False otherwise.
        """
        ...

    def copy(self, stack: bool, deep_copy_legal_moves: bool = True) -> Self:
        """
        Create a copy of the current board.

        Args:
            stack (bool): Whether to copy the previous action stack as well. Important in some games.
            deep_copy_legal_moves (bool): Whether to deep copy the legal moves generator.

        Returns:
            BoardChi: A new instance of the BoardChi class with the copied board.
        """
        ...

    def step(self, branch_key: BranchKey) -> StateModifications | None:
        """Advances the state by applying the action corresponding to the given branch key.

        Args:
            branch_key (BranchKey): The branch key representing the action to be applied.

        Returns:
            StateModifications | None: The modifications to the state after applying the action, or None if the action is invalid.
        """
        ...

    def pprint(self) -> str:
        """Returns a pretty-printed string representation of the content.

        Returns:
            str: A pretty-printed string representation of the content.
        """
        ...


type ColorIndex = Annotated[int, "1 for white, 0 for black"]

WHITE: ColorIndex = 1
BLACK: ColorIndex = 0


class Color(int, Enum):
    """Represents the color of a player in a game."""

    WHITE = WHITE
    BLACK = BLACK


def _actions_history_factory() -> list[ActionKey]:
    return []


@dataclass
class StatePlusHistory[StateT]:
    """A State with historical actions and states."""

    @staticmethod
    def _states_factory() -> list[StateT]:
        return []

    current_state_tag: StateTag
    historical_actions: list[ActionKey] = field(
        default_factory=_actions_history_factory
    )
    historical_states: list[StateT] = field(default_factory=_states_factory)


class HasTurn(Protocol):
    """Protocol for a content object that has a tag."""

    @property
    def turn(self) -> Color:
        """Returns the tag of the content.

        Returns:
            ContentTag: The tag of the content.
        """
        ...


class TurnState(State, HasTurn, Protocol):
    """A State that also supports turn()."""

    ...


@dataclass
class TurnStatePlusHistory[StateT = Any]:
    """A TurnState with historical actions and states."""

    @staticmethod
    def _states_factory() -> list[StateT]:
        return []

    current_state_tag: StateTag
    turn: Color
    historical_actions: list[ActionKey] = field(
        default_factory=_actions_history_factory
    )
    historical_states: list[StateT] = field(default_factory=_states_factory)



import pytest

from valanga.game import Color
from valanga.over_event import HowOver, OverEvent, OverTags, Winner


@pytest.mark.parametrize(
    "who, expected",
    [
        (Winner.WHITE, OverTags.TAG_WIN_WHITE),
        (Winner.BLACK, OverTags.TAG_WIN_BLACK),
    ],
)
def test_get_over_tag_win_outcomes(who, expected):
    event = OverEvent(how_over=HowOver.WIN, who_is_winner=who)

    assert event.get_over_tag() == expected
    assert event.is_over() is True
    assert event.is_win() is True
    assert event.is_draw() is False


@pytest.mark.parametrize("winner", [Winner.WHITE, Winner.BLACK])
def test_is_winner_matches_color(winner):
    event = OverEvent(how_over=HowOver.WIN, who_is_winner=winner)

    assert event.is_winner(Color.WHITE) is (winner == Winner.WHITE)
    assert event.is_winner(Color.BLACK) is (winner == Winner.BLACK)


@pytest.mark.parametrize(
    "how_over, winner, expected",
    [
        (HowOver.DRAW, Winner.NO_KNOWN_WINNER, OverTags.TAG_DRAW),
        (HowOver.DO_NOT_KNOW_OVER, Winner.NO_KNOWN_WINNER, OverTags.TAG_DO_NOT_KNOW),
    ],
)
def test_get_over_tag_draw_and_unknown(how_over, winner, expected):
    event = OverEvent(how_over=how_over, who_is_winner=winner)

    assert event.get_over_tag() == expected
    assert event.is_over() is (how_over == HowOver.DRAW)
    assert event.is_win() is False
    assert event.is_draw() is (how_over == HowOver.DRAW)


@pytest.mark.parametrize(
    "how_over, who_is_winner",
    [
        (HowOver.WIN, Winner.NO_KNOWN_WINNER),
        (HowOver.DRAW, Winner.WHITE),
    ],
)
def test_post_init_validates_winner_configuration(how_over, who_is_winner):
    with pytest.raises(AssertionError):
        OverEvent(how_over=how_over, who_is_winner=who_is_winner)


@pytest.mark.parametrize("player", [Color.WHITE, Color.BLACK])
def test_is_winner_requires_win(player):
    event = OverEvent(how_over=HowOver.DRAW, who_is_winner=Winner.NO_KNOWN_WINNER)

    assert event.is_winner(player) is False


@pytest.mark.parametrize(
    "how_over, expected_exception",
    [
        (HowOver.WIN, ValueError),
        (None, ValueError),
    ],
)
def test_get_over_tag_invalid_configurations(how_over, expected_exception):
    event = OverEvent(how_over=HowOver.WIN, who_is_winner=Winner.WHITE)
    event.how_over = how_over  # mutate to bypass post-init validation
    event.who_is_winner = Winner.NO_KNOWN_WINNER

    with pytest.raises(expected_exception):
        event.get_over_tag()


@pytest.mark.parametrize(
    "how_over, who_is_winner",
    [
        (HowOver.WIN, Winner.WHITE),
        (HowOver.WIN, Winner.BLACK),
    ],
)
def test_test_method_for_wins(how_over, who_is_winner):
    event = OverEvent(how_over=how_over, who_is_winner=who_is_winner)

    event.test()


def test_test_method_invalid_draw_configuration():
    event = OverEvent(how_over=HowOver.DRAW, who_is_winner=Winner.NO_KNOWN_WINNER)

    with pytest.raises(AssertionError):
        event.test()


def test_bool_raises():
    event = OverEvent()

    with pytest.raises(ValueError):
        bool(event)



"""
Contains the definition of the ContentRepresentation protocol for content representations used in evaluations.
"""

from typing import Protocol, TypeVar

from valanga.evaluator_types import EvaluatorInput

from .game import State

StateT_contra = TypeVar("StateT_contra", bound=State, contravariant=True, default=State)

EvalIn_co = TypeVar(
    "EvalIn_co", covariant=True, bound=EvaluatorInput, default=EvaluatorInput
)


class ContentRepresentation[StateT_contra, EvalIn_co](Protocol):
    """
    Protocol defining the interface for a content representation.
    It is a function returning the proper input for evaluation by the content evaluator.
    """

    def get_evaluator_input(self, state: StateT_contra) -> EvalIn_co:
        """
        Returns the evaluator input tensor for the content. Content representations have generally a compressed view and complemetary view of state info so to avoid redundancy and have all the necessary info we also give the state as input.

        Args:
            state: The current state of the game.

        Returns:
            The evaluator input tensor.
        """
        ...



"""
Policy-related classes and protocols for branch selection in game trees.
"""

from dataclasses import dataclass
from typing import Mapping, Protocol

from valanga.evaluations import StateEvaluation
from valanga.game import BranchKey, BranchName, Seed, State


@dataclass(frozen=True, slots=True)
class BranchPolicy:
    """
    Represents a probability distribution over branches.
    """

    probs: Mapping[BranchKey, float]  # should sum to ~1.0


@dataclass(frozen=True, slots=True)
class Recommendation:
    """A recommendation for a specific branch in a tree node."""

    recommended_name: BranchName
    evaluation: StateEvaluation | None = None
    policy: BranchPolicy | None = None
    branch_evals: Mapping[BranchName, StateEvaluation] | None = None


class BranchSelector(Protocol):
    """Protocol for a branch selector."""

    def recommend(self, state: State, seed: Seed) -> Recommendation:
        """Given a state and a seed, recommends a branch to take.
        Args:
            state (State): The current state of the game.
            seed (Seed): A seed for any randomness involved in the selection.
        Returns:
            Recommendation: The recommended branch to take.
        """
        ...



"""Common types and utilities shared by multiple libraries."""

from .evaluations import EvalItem, FloatyStateEvaluation, ForcedOutcome, StateEvaluation
from .game import (
    BLACK,
    WHITE,
    BranchKey,
    BranchKeyGeneratorP,
    Color,
    ColorIndex,
    HasTurn,
    State,
    StateModifications,
    StateTag,
    TurnState,
)
from .over_event import OverEvent
from .progress_messsage import PlayerProgressMessage
from .representation_factory import RepresentationFactory
from .represention_for_evaluation import ContentRepresentation

__all__ = [
    "ForcedOutcome",
    "FloatyStateEvaluation",
    "StateEvaluation",
    "EvalItem",
    "OverEvent",
    "Color",
    "WHITE",
    "BLACK",
    "ColorIndex",
    "HasTurn",
    "TurnState",
    "ContentRepresentation",
    "RepresentationFactory",
    "State",
    "StateModifications",
    "BranchKeyGeneratorP",
    "BranchKey",
    "PlayerProgressMessage",
    "StateTag",
]



"""
Evaluation-related classes and types.
"""

from dataclasses import dataclass
from typing import Protocol

from valanga.evaluator_types import EvaluatorInput

from .game import BranchKey, State
from .over_event import OverEvent
from .represention_for_evaluation import ContentRepresentation


class EvalItem[StateT: State](Protocol):
    """
    The protocol for an evaluation item.
    An evaluation item is something that has a state and optionally a representation of that state.
    """

    @property
    def state(self) -> StateT:
        """The state associated with this evaluation item."""
        ...

    @property
    def state_representation(
        self,
    ) -> ContentRepresentation[StateT, EvaluatorInput] | None:
        """The representation of the state associated with this evaluation item, if available."""
        ...


@dataclass
class ForcedOutcome:
    """
    The class
    """

    # The forced outcome with optimal play by both sides.
    outcome: OverEvent

    # the line
    line: list[BranchKey]


@dataclass
class FloatyStateEvaluation:
    """
    The class to defines what is an evaluation of a board.
    By convention is it always evaluated from the view point of the white side.
    """

    # The evaluation value for the white side when the outcome is not certain. Typically, a float.
    # todo can we remove the None option?
    value_white: float | None


StateEvaluation = FloatyStateEvaluation | ForcedOutcome
